/*
 +---------------------------------------------------------------------+
 | ups.cc                                                              |
 | ups class implementation for APC SmartUPS 700                          |
 |                                                                     |
 | Author:  Bob Pill, NCIS                                             |
 |          bobp@newcorridor.com                                       |
 |                                                                     |
 | Created: Sept 15, 1997                                              |
 |                                                                     |
 | Copyright (c) 1997 New Corridor Information Services.               |
 | All rights reserved.                                                |
 |                                                                     |
 |                                                                     |
 | THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR        |
 | IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED      |
 | WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR          |
 | PURPOSE.                                                            |
 |                                                                     |
 |---------------------------------------------------------------------|
 |  ***Revision-History***  Begin-Log                                  |
 |     Initial creation - (started at ver 2.0.1) Sept 15, 1997 (BP)    |
 |     Version 2.1.0 update Jan 01, 1998 (BP)                          |
 |     Version 2.1.1 update Feb 03, 1998 (BP)                          |
 |     Version 2.1.2 update May 17, 1998 (BP)                          |
 |     Version 2.1.3 update Aug 04, 1998 (BP)                          |
 |  ***Revision-History***  End-Log                                    |
 +---------------------------------------------------------------------+
*/

#include <sys/types.h>
#include <sys/time.h>
#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>
#include <termios.h>
#include <string.h>
#include <memory.h>
#include "stddef.h"
#include "ups.h"


// constructor
Ups::Ups(const char* pszTTYName)
{
    m_fdPort = -1;
    m_iDataLen = 0;
    memset(&m_Status,0,sizeof(m_Status));
    m_Status.byNotAvailable = TRUE;
    memset(m_szResponse,0,sizeof(m_szResponse));
    memset(m_szTTYName,0,sizeof(m_szTTYName));
    strcpy(m_szTTYName,pszTTYName);
}


// destructor
Ups::~Ups(void)
{
    if(m_fdPort != -1)
    {
      Reset();
      close(m_fdPort);
    }
}


int Ups::Init(void)
{
    int iRetCode = RC_SUCCESS;
    char szError[ERRORMSG_SIZE];

    if((m_fdPort = open(m_szTTYName, O_RDWR|O_EXCL|O_NONBLOCK, 0)) == -1)
    {
      sprintf(szError,"\nopen tty %s",m_szTTYName);
      perror(szError);
      iRetCode = RC_FILE_OPEN_ERROR;
      goto InitExit;
    }

    // turn on raw mode on UPS tty line
    if(tcgetattr(m_fdPort, &m_oMode) == -1)
    {
      perror("\ntcgetattr tty: ");
      iRetCode = RC_DEVICE_CONTROL_ERROR;
    }

    m_upsMode = m_oMode;
    cfmakeraw(&m_upsMode);
    cfsetispeed(&m_upsMode, B2400);
    cfsetospeed(&m_upsMode, B2400);  
    m_upsMode.c_cflag |= CLOCAL;
    if(tcsetattr(m_fdPort, TCSANOW, &m_upsMode) == -1)
    {
      perror("\ntcsetattr RAW: ");
      iRetCode = RC_DEVICE_CONTROL_ERROR;
      goto InitExit;
    }

    // set UPS to smart mode
    Command(UPS_CMD_SMART);

InitExit:

    return(iRetCode);
}


void Ups::Reset(void)
{
  // UPS back to dumb mode
  Command(UPS_CMD_DUMB);
  // tty line back to normal mode
  tcsetattr(m_fdPort, TCSANOW, &m_oMode);
}



// send single byte command to UPS.  Attempt to read response
// string from UPS.  Read up terminating CR-LF and store result in
// m_szResponse.
void Ups::Command(BYTE byCommand)
{
    char cByte;
    fd_set fdsRead;
    struct timeval timeOut;
  
    memset(m_szResponse,0,sizeof(m_szResponse));
    if(write(m_fdPort, &byCommand, 1) < 1)
      return;

    for(m_iDataLen = 0; m_iDataLen < MAX_UPS_RESPONSE - 1; m_iDataLen++)
    {
      FD_ZERO(&fdsRead);
      FD_SET(m_fdPort, &fdsRead);
      timeOut.tv_sec = 0;
      timeOut.tv_usec = 100 * 1000;
      if(select(FD_SETSIZE, &fdsRead, 0, 0, &timeOut) < 1)
        break;
      if(read(m_fdPort, &cByte, 1) < 1)
        break;
      if(cByte == '\r')
      {
        m_iDataLen--;
        continue;
      }
      if(cByte == '\n')
        break;
      m_szResponse[m_iDataLen] = cByte;
    }
}


const UpsStatus& Ups::ReadStatus(void)
{
    int iStatus = 0;

    m_Status.byStandby = FALSE;
    m_Status.byPowerFailure = FALSE;
    m_Status.byBatteryLow = FALSE;
    m_Status.byNotAvailable = TRUE;
      
    Command(UPS_CMD_STATUS);
    if(m_iDataLen == 0)
      goto ReadStatusExit;

    m_Status.byNotAvailable = FALSE;
    sscanf(m_szResponse, "%2x", &iStatus);
    if(iStatus & UPS_STATUS_BLOW)
      m_Status.byBatteryLow = TRUE;
    if(iStatus & UPS_STATUS_ON)
      goto ReadStatusExit;
    if(iStatus & UPS_STATUS_PWRF)
      m_Status.byPowerFailure = TRUE;
    else if(iStatus & UPS_STATUS_STBY)
    {
      m_Status.byPowerFailure = TRUE;
      m_Status.byStandby = TRUE;
    }
    else
      m_Status.byStandby = TRUE;

ReadStatusExit:

    return m_Status;
}



const char* Ups::ReadString(BYTE byCommand)
{
    const char* pString = NULL;

    Command(byCommand);
    if(m_iDataLen == 0 || strcmp(m_szResponse,UPS_RES_NA) == 0)
      goto ReadStringExit;
    else
      pString = m_szResponse;

ReadStringExit:

    return pString;
}



double Ups::ReadDouble(BYTE byCommand)
{
    double dValue = 0.0;

    Command(byCommand);
    if(m_iDataLen == 0 || strcmp(m_szResponse,UPS_RES_NA) == 0)
      goto ReadDoubleExit;
    else
      dValue = atof(m_szResponse);

ReadDoubleExit:

    return dValue;
}


long Ups::ReadLong(BYTE byCommand)
{
    long iValue = 0;

    Command(byCommand);
    if(m_iDataLen == 0 || strcmp(m_szResponse,UPS_RES_NA) == 0)
      goto ReadLongExit;
    else
      iValue = atol(m_szResponse);

ReadLongExit:

    return iValue;
}


char Ups::ReadEvent(void)
{
    char cEvent = 0;

    read(m_fdPort, &cEvent, 1);
    return cEvent;
}


void Ups::Shutdown(void)
{
    Command(UPS_CMD_SHUT_SOFT);
    sleep(1);	
    Command(UPS_CMD_SHUT_SOFT);
    sleep(1);
    Command(ASCII_CR);
}

//---------------------------------------------------------------------
// end-of-file: ups.cc                                                 
//---------------------------------------------------------------------
