/*
 +---------------------------------------------------------------------+
 | upsmonitor.cc                                                       |
 | ups class implementation for APC smartups                           |
 |                                                                     |
 | Author:  Bob Pill, NCIS                                             |
 |          bobp@newcorridor.com                                       |
 |                                                                     |
 | Created: Sept 15, 1997                                              |
 |                                                                     |
 | Copyright (c) 1997 New Corridor Information Services.               |
 | All rights reserved.                                                |
 |                                                                     |
 |                                                                     |
 | THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR        |
 | IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED      |
 | WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR          |
 | PURPOSE.                                                            |
 |                                                                     |
 |---------------------------------------------------------------------|
 |  ***Revision-History***  Begin-Log                                  |
 |     Initial creation - (started at ver 2.0.1) Sept 15, 1997 (BP)    |
 |     Version 2.1.0 update Jan 01, 1998 (BP)                          |
 |     Version 2.1.1 update Feb 03, 1998 (BP)                          |
 |     Version 2.1.2 update May 17, 1998 (BP)                          |
 |     Version 2.1.3 update Aug 04, 1998 (BP)                          |
 |  ***Revision-History***  End-Log                                    |
 +---------------------------------------------------------------------+
*/

#include <sys/types.h>
#include <sys/time.h>
#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>
#include <termios.h>
#include <string.h>
#include <memory.h>
#include <syslog.h>

#include "stddef.h"
#include "upsmonitor.h"


// constructor
UpsMonitor::UpsMonitor(const char* pszTTYName, const char* pszSocketName, int iRetryLimit, const char* pszUpsmondVersion)
{
    m_iRetryLimit = iRetryLimit * 60;  // number of seconds before shutdown after power failure
    m_pUps = new Ups(pszTTYName);
    m_pSocket = new CSocket(pszSocketName);
    m_iState = UPS_IDLE;
    m_tLastSelftest = 0;
    m_bBatteryLow = FALSE;
    m_bClientConnected = FALSE;
    m_tTimePowerFail = 0;
    strcpy(m_szUpsmondVersion,pszUpsmondVersion);
}


// destructor
UpsMonitor::~UpsMonitor(void)
{
    if(m_pUps)
      delete m_pUps;

    if(m_pSocket)
      delete m_pSocket;
}


int UpsMonitor::Init(void)
{
    int iRetCode = RC_SUCCESS;

    if(m_pUps && m_pSocket)
    {
      iRetCode = m_pSocket->Listen();
      iRetCode = m_pUps->Init();
    }
    else
      iRetCode = RC_MEMORY_ALLOCATION_ERROR;

    return(iRetCode);
}


void UpsMonitor::HandleUPSEvent(void)
{
    char cEvent;

    cEvent = m_pUps->ReadEvent();
  
    switch(cEvent)
    {
      case UPS_EV_TURNON:
        m_tLastSelftest = time(NULL);
        m_pUps->Command(UPS_CMD_SMART);
        syslog(LOG_INFO,"UPS device turned on.");
        break;
      
      case UPS_EV_POWER:
        if(m_bBatteryLow)
          syslog(LOG_CRIT,"Input line power failure, battery low");
        else
        {
          syslog(LOG_CRIT,"Input line power failure");
          if(m_iState == UPS_IDLE)
            m_iState = UPS_POWER_FAIL_START;
        }          
        break;

      case UPS_EV_PWRBACK:
        syslog(LOG_ALERT,"Input line power back");
        m_iState = UPS_IDLE;
        break;

      case UPS_EV_BATLOW:
        // don't warn if this is immediately after a selftest/turnon
        if(time(NULL) - m_tLastSelftest < 5)
          break;
        m_bBatteryLow = TRUE;
        syslog(LOG_CRIT,"Battery low");
        break;

      case UPS_EV_BATUP:
        m_bBatteryLow = FALSE;
        m_tLastSelftest = 0;
        syslog(LOG_ALERT,"Battery up again");
      break;

      default:
        syslog(LOG_ALERT,"Unhandled event %c",cEvent);
        break;
    }
}



int UpsMonitor::GetSocket(void)
{
    return(m_bClientConnected 
           ? m_pSocket->GetPartnerSocket()
           : m_pSocket->GetListenSocket()
           );
}



int UpsMonitor::PostErrorResponse(int iRetCode, int iFunctionID)
{
    UpsInfoResp *pResp;       // ups info response packet
    CBuffer *pbufSend;        // socket send buffer

    pbufSend = m_pSocket->GetSendBuffer();
    pResp = (UpsInfoResp *)(pbufSend->GetBuffer());
    memset(pResp,0,sizeof(UpsInfoResp));

    pResp->iMsgFormatID = UPSInfoMsgFormatID;
    pResp->iFunctionID = iFunctionID;
    pResp->iRetCode = iRetCode;
    pbufSend->SetLength(sizeof(UpsInfoResp));
    return(m_pSocket->Write());
}




int UpsMonitor::PostNormalResponse(void)
{
    UpsStatus upsStatus;      // ups status
    UpsInfoResp *pResp;       // ups info response packet
    CBuffer *pbufSend;        // socket send buffer

    pbufSend = m_pSocket->GetSendBuffer();
    pResp = (UpsInfoResp *)(pbufSend->GetBuffer());
    memset(pResp,0,sizeof(UpsInfoResp));

    upsStatus = m_pUps->ReadStatus();
    pResp->iMsgFormatID = UPSInfoMsgFormatID;
    pResp->iFunctionID = UPS_GET_STATUS;
    pResp->iRetCode = RC_SUCCESS;
    strcpy(pResp->szUpsmondVersion,m_szUpsmondVersion);
    strcpy(pResp->szModel,m_pUps->ReadString(UPS_CMD_MODEL));
    strcpy(pResp->szCopyRight,m_pUps->ReadString(UPS_CMD_COPYRIGHT));
    strcpy(pResp->szMfgDate,m_pUps->ReadString(UPS_CMD_MDATE));
    strcpy(pResp->szSerial,m_pUps->ReadString(UPS_CMD_ID));
    strcpy(pResp->szLocalID,m_pUps->ReadString(UPS_CMD_LOCALID));
    strcpy(pResp->szFirmware,m_pUps->ReadString(UPS_CMD_FIRMWARE));
    strcpy(pResp->szSmartUPSVersion,m_pUps->ReadString(UPS_CMD_VERS));
    strcpy(pResp->szBattReplaceDate,m_pUps->ReadString(UPS_CMD_REPLDATE));
    pResp->iNomLine = atoi(m_pUps->ReadString(UPS_CMD_NLINE));
    pResp->iNomBatt = (int)(m_pUps->ReadLong(UPS_CMD_NBATTERY));
    pResp->iAvailable = upsStatus.byNotAvailable ? FALSE : TRUE;
    pResp->iStandby = upsStatus.byStandby;
    pResp->iLine = upsStatus.byPowerFailure ? FALSE : TRUE;
    pResp->iLowBatt = upsStatus.byBatteryLow;
    pResp->dLineVoltage = m_pUps->ReadDouble(UPS_CMD_LINE);
    pResp->dBattVoltage = m_pUps->ReadDouble(UPS_CMD_BATTERY);
    pResp->dLoad = m_pUps->ReadDouble(UPS_CMD_POWER);
    pResp->dBatteryLevel = m_pUps->ReadDouble(UPS_CMD_LEVEL);
    pbufSend->SetLength(sizeof(UpsInfoResp));
    return(m_pSocket->Write());
}



int UpsMonitor::ProcessRequest(UpsInfoReq *pReq)
{
    int iRetCode = RC_SUCCESS;
    int iErrorCode = RC_SUCCESS;

    if(pReq->iMsgFormatID == UPSInfoMsgFormatID)
    {
      switch(pReq->iFunctionID)
      {
        case UPS_GET_STATUS:
          //  only client request supported for now
          iRetCode = PostNormalResponse();
          break;

        default:
          iErrorCode = RC_UNKNOWN_FUNCTION_REQUEST;
          break;
      }
    }
    else
      iErrorCode = RC_FORMAT_ERROR;

    if(iErrorCode)
      iRetCode = PostErrorResponse(iErrorCode,pReq->iFunctionID);

    return(iRetCode);
}



void UpsMonitor::HandleSocketEvent(void)
{
    int iRetCode = RC_SUCCESS;
    CBuffer *pbufRcve = NULL;
    UpsInfoReq *pReq = NULL;

    if(m_bClientConnected)
    {
      // if we are connected to client, then socket event
      // is a client request
      iRetCode = m_pSocket->Read();
      if(iRetCode == RC_SOCKET_READ_ERROR)
      {
        m_bClientConnected = FALSE;
        m_pSocket->Release();
      }
      else
      {
        pbufRcve = m_pSocket->GetRcveBuffer();
        if(pbufRcve->GetLength() == 0)
        {
          // no data means client disconnected, so we revert
          // back to listening on server socket for new connection
          m_bClientConnected = FALSE;
          m_pSocket->Release();
        }
        else // process client requests 
        {
          pReq = (UpsInfoReq *)pbufRcve->GetBuffer();
          iRetCode = ProcessRequest(pReq);
        }
      }
    }
    else // socket event is a connection
    {
      m_pSocket->Accept();
      m_bClientConnected = TRUE;
    }
}



int UpsMonitor::Run(void)
{
    int iExitMode = UPSMONITOR_EXIT_NORMAL;
    time_t tElapseTime = 0;
    BOOL bRun = TRUE;
    int fdSocket;
    fd_set fdsMonitor;
    struct timeval timeOut;

#if defined(DEBUG)
    int iDebugCount = 0;
#endif

    timeOut.tv_sec = MONITOR_SLEEP_INTERVAL;
    timeOut.tv_usec = 0;

    while(bRun)
    {
      FD_ZERO(&fdsMonitor);
      FD_SET(m_pUps->GetPortFD(), &fdsMonitor);

      fdSocket = GetSocket();
      FD_SET(fdSocket, &fdsMonitor);

      // block waiting for event on UPS or Socket
      if(select(FD_SETSIZE, &fdsMonitor, 0, 0, &timeOut) < 0)
        continue;

#if defined(DEBUG)
      ++iDebugCount;
#endif
      

      // did Socket event fire
      if(FD_ISSET(fdSocket, &fdsMonitor))
      {
        HandleSocketEvent();

#if defined(DEBUG)
      syslog(LOG_INFO,"Socket Event: Cycle: %d  State: %d",iDebugCount,m_iState);
#endif        
      }

      // did UPS event fire
      if(FD_ISSET(m_pUps->GetPortFD(), &fdsMonitor))
      {
        HandleUPSEvent();

#if defined(DEBUG)
      syslog(LOG_INFO,"UPS Event: Cycle: %d  State: %d",iDebugCount,m_iState);
#endif        
      }

      // The state machine times how long power stays down and        
      // will result in system shutdown after 'm_iRetryLimit' seconds.
      // The 'm_iRetryLimit' is a command line option and defaults to 
      // 300 seconds (5 minutes).  This 5 minute default means the    
      // power must stay down for 5 minutes before we do a shutdown.  
      // If the power comes back up at any time during the retry      
      // interval, everything is reset and we start counting from the 
      // top the next time the power drops.                           

      switch(m_iState)
      {
        case UPS_IDLE:
          break;

        case UPS_POWER_FAIL_START:
          m_tTimePowerFail = time(NULL);
          m_iState = UPS_SHUTDOWN_PENDING;
          break;

        case UPS_SHUTDOWN_PENDING:
          tElapseTime = time(NULL) - m_tTimePowerFail;
          if(tElapseTime >= m_iRetryLimit)
          {
            // inititiate system shutdown when retry limit expires
            syslog(LOG_EMERG,"Power Failure - System shutdown now!");
            m_iState = UPS_SHUTDOWN_FINAL;
            iExitMode = UPSMONITOR_EXIT_SHUTDOWN;
            bRun = FALSE;
          }
          else if(tElapseTime >= 60)
            // begin shutdown warnings only after power has been down
            // for at least 60 seconds
            syslog(LOG_EMERG,"Power Failure - %d seconds to shutdown!",(m_iRetryLimit - tElapseTime));
          break;

      } // end switch
    } // end while

    return(iExitMode);
}



void UpsMonitor::ShutdownUPS(void)
{
    if(m_pUps)
      m_pUps->Shutdown();
}


//---------------------------------------------------------------------
// end-of-file: upsmonitor.cc                                          
//---------------------------------------------------------------------
